/* acl.c - access control lists

   Copyright (C) 2002, 2003, 2005 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

   Written by Paul Eggert.  */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#if ENABLE_NLS
# include <libintl.h>
# define _(Text) gettext (Text)
#else
# define _(Text) Text
#endif

#include <sys/types.h>
#include <sys/stat.h>
#ifndef S_ISLNK
# define S_ISLNK(Mode) 0
#endif

#include <stdlib.h>
#include <acl/libacl.h>

#include "acl.h"
#include "error.h"
#include "quote.h"

#include <errno.h>
#ifndef ENOSYS
# define ENOSYS (-1)
#endif
#ifndef ENOTSUP
# define ENOTSUP (-1)
#endif

#ifndef MIN_ACL_ENTRIES
# define MIN_ACL_ENTRIES 4
#endif

/* Return 1 if FILE has a nontrivial access control list, 0 if not,
   and -1 (setting errno) if an error is encountered.  */

int
file_has_acl (char const *file, struct stat const *filestat)
{
#if HAVE_ACL && defined GETACLCNT
  /* This implementation should work on recent-enough versions of HP-UX,
     Solaris, and Unixware.  */

  if (! S_ISLNK (filestat->st_mode))
    {
      int n = acl (file, GETACLCNT, 0, NULL);
      return n < 0 ? (errno == ENOSYS ? 0 : -1) : (MIN_ACL_ENTRIES < n);
    }
#elif HAVE_ACL_EXTENDED_FILE

  /* Linux specific.  */

  if (! S_ISLNK (filestat->st_mode))
    {
      int ret = acl_extended_file (file);
      if (ret < 0)
	return (errno == ENOSYS || errno == ENOTSUP) ? 0 : -1;
      return ret;
    }
#else
  /* FIXME: Add support for AIX, Irix, and Tru64, FreeBSD, etc.
     Please see Samba's source/lib/sysacls.c file for fix-related ideas.  */
#endif

  return 0;
}

/* Copy the permissions from SRC_PATH to DST_PATH, including access control
   lists on systems where this is supported. MODE is the file mode for
   DST_PATH, including the file type.
   Also sets special bits in MODE on DST_PATH.  */

int
copy_acl (char const *src_path, char const *dst_path, mode_t mode)
{
#if HAVE_ACL_GET_FILE && HAVE_ACL_SET_FILE && HAVE_ACL_FREE && \
    HAVE_ACL_ENTRIES

  /* Linux specific. Will work on all POSIX 1003.1e draft 17 (abandoned)
     compliant systems if the acl_entries() function is implemented.  */

  acl_t acl = acl_get_file (src_path, ACL_TYPE_ACCESS);
  if (acl == NULL)
    {
      if (errno == ENOSYS || errno == ENOTSUP)
	return set_acl (dst_path, mode);
      else
        {
	  error (0, errno, "%s", quote (src_path));
	  return -1;
	}
    }

  if (acl_set_file (dst_path, ACL_TYPE_ACCESS, acl))
    {
      int saved_errno = errno;

      if (errno == ENOSYS || errno == ENOTSUP)
        {
	  int n = acl_entries (acl);

	  acl_free (acl);
	  if (n == 3)
	    {
	      if (chmod (dst_path, mode))
		saved_errno = errno;
	      else
		return 0;
	    }
	  else
	    chmod (dst_path, mode);
	}
      else
	{
	  acl_free (acl);
	  chmod (dst_path, mode);
	}
      error (0, saved_errno, _("preserving permissions for %s"),
	     quote (dst_path));
      return -1;
    }
  acl_free (acl);

  if (mode & (S_ISUID | S_ISGID | S_ISVTX))
    {
      /* We did not call chmod so far, so the special bits have not yet
         been set.  */

      if (chmod (dst_path, mode))
	{
	  error (0, errno, _("preserving permissions for %s"),
	  	 quote (dst_path));
	  return -1;
	}
    }

  if (S_ISDIR (mode))
    {
      acl = acl_get_file (src_path, ACL_TYPE_DEFAULT);
      if (acl == NULL)
	{
	  error (0, errno, "%s", quote (src_path));
	  return -1;
	}

      if (acl_set_file (dst_path, ACL_TYPE_DEFAULT, acl))
	{
	  error (0, errno, _("preserving permissions for %s"),
		 quote (dst_path));
	  acl_free(acl);
	  return -1;
	}
      else
        acl_free(acl);
    }
  return 0;
#else
  int ret = chmod (dst_path, mode);
  if (ret)
    error (0, errno, _("preserving permissions for %s"), quote (dst_path));
  return ret;
#endif
}

/* Set the permissions of PATH, overwriting access control lists, on systems
   where this is supported. MODE is the file mode for PATH, including the
   file type. Also sets special bits in MODE on PATH.  */

int
set_acl (char const *path, mode_t mode)
{
#if HAVE_ACL_FROM_TEXT && HAVE_ACL_SET_FILE && HAVE_ACL_FREE && \
       HAVE_ACL_DELETE_DEF_FILE
  char acl_text[] = "u::---,g::---,o::---";
  acl_t acl;

  if (mode & S_IRUSR) acl_text[ 3] = 'r';
  if (mode & S_IWUSR) acl_text[ 4] = 'w';
  if (mode & S_IXUSR) acl_text[ 5] = 'x';
  if (mode & S_IRGRP) acl_text[10] = 'r';
  if (mode & S_IWGRP) acl_text[11] = 'w';
  if (mode & S_IXGRP) acl_text[12] = 'x';
  if (mode & S_IROTH) acl_text[17] = 'r';
  if (mode & S_IWOTH) acl_text[18] = 'w';
  if (mode & S_IXOTH) acl_text[19] = 'x';

  acl = acl_from_text(acl_text);
  if (!acl)
    {
      error (0, errno, "%s", quote (path));
      return -1;
    }

  if (acl_set_file(path, ACL_TYPE_ACCESS, acl))
    {
      int saved_errno = errno;
      acl_free (acl);

      if (errno == ENOTSUP || errno == ENOSYS)
	{
	  if (chmod (path, mode))
	    saved_errno = errno;
	  else
	    return 0;
	}
      error (0, saved_errno, _("setting permissions for %s"), quote (path));
      return -1;
    }
  acl_free (acl);

  if (mode & (S_ISUID | S_ISGID | S_ISVTX))
    {
      /* We did not call chmod so far, so the special bits have not yet
         been set.  */

      if (chmod (path, mode))
	{
	  error (0, errno, _("preserving permissions for %s"),
	  	 quote (path));
	  return -1;
	}
    }

  if (S_ISDIR (mode) && acl_delete_def_file (path))
    {
      error (0, errno, _("setting permissions for %s"), quote (path));
      return -1;
    }
  return 0;
#else
  int ret = chmod (path, mode);
  if (ret)
    error (0, errno, _("setting permissions for %s"), quote (path));
  return ret;
#endif
}

