/* chcontext -- change security context of a pathname */

#include <config.h>
#include <stdio.h>
#include <sys/types.h>
#include <grp.h>
#include <getopt.h>
#include <selinux/selinux.h>
#include <selinux/context.h>

#include "system.h"
#include "error.h"
#include "savedir.h"
#include "group-member.h"

enum Change_status
{
  CH_SUCCEEDED,
  CH_FAILED,
  CH_NO_CHANGE_REQUESTED
};

enum Verbosity
{
  /* Print a message for each file that is processed.  */
  V_high,

  /* Print a message for each file whose attributes we change.  */
  V_changes_only,

  /* Do not be verbose.  This is the default. */
  V_off
};

static int change_dir_context (const char *dir, const struct stat *statp);

/* The name the program was run with. */
char *program_name;

/* If nonzero, and the systems has support for it, change the context
   of symbolic links rather than any files they point to.  */
static int change_symlinks;

/* If nonzero, change the context of directories recursively. */
static int recurse;

/* If nonzero, force silence (no error messages). */
static int force_silent;

/* Level of verbosity.  */
static enum Verbosity verbosity = V_off;

/* The name of the context file is being given. */
static const char *specified_context;

/* Specific components of the context */
static const char *specified_user;
static const char *specified_role;
static const char *specified_range;
static const char *specified_type;

/* The argument to the --reference option.  Use the context of this file.
   This file must exist.  */
static char *reference_file;

/* If nonzero, display usage information and exit.  */
static int show_help;

/* If nonzero, print the version on standard output and exit.  */
static int show_version;

static struct option const long_options[] =
{
  {"recursive", no_argument, 0, 'R'},
  {"changes", no_argument, 0, 'c'},
  {"no-dereference", no_argument, 0, 'h'},
  {"silent", no_argument, 0, 'f'},
  {"quiet", no_argument, 0, 'f'},
  {"reference", required_argument, 0, CHAR_MAX + 1},
  {"context", required_argument, 0, CHAR_MAX + 2},
  {"user", required_argument, 0, 'u'},
  {"role", required_argument, 0, 'r'},      
  {"type", required_argument, 0, 't'},
  {"range", required_argument, 0, 'l'},   
  {"verbose", no_argument, 0, 'v'},
  {"help", no_argument, &show_help, 1},
  {"version", no_argument, &show_version, 1},
  {0, 0, 0, 0}
};

/* Tell the user how/if the context of FILE has been changed.
   CHANGED describes what (if anything) has happened. */

static void
describe_change (const char *file, security_context_t newcontext, enum Change_status changed)
{
  const char *fmt;
  switch (changed)
    {
    case CH_SUCCEEDED:
      fmt = _("context of %s changed to %s\n");
      break;
    case CH_FAILED:
      fmt = _("failed to change context of %s to %s\n");
      break;
    case CH_NO_CHANGE_REQUESTED:
      fmt = _("context of %s retained as %s\n");
      break;
    default:
      abort ();
    }
  printf (fmt, file, newcontext);
}

static int
compute_context_from_mask (security_context_t context, context_t *ret)
{
  context_t newcontext = context_new (context);
  if (!newcontext)
    return 1;
#define SETCOMPONENT(comp) \
  do { \
    if (specified_ ## comp) \
      if (context_ ## comp ## _set (newcontext, specified_ ## comp)) \
        goto lose; \
  } while (0)
      
  SETCOMPONENT(user);
  SETCOMPONENT(range);
  SETCOMPONENT(role);
  SETCOMPONENT(type);
#undef SETCOMPONENT
  
  *ret = newcontext;
  return 0;
 lose:
  context_free (newcontext);
  return 1;
}

/* Change the context of FILE, using specified components.
   If it is a directory and -R is given, recurse.
   Return 0 if successful, 1 if errors occurred. */

static int
change_file_context (const char *file)
{
  struct stat file_stats;
  security_context_t file_context=NULL;
  context_t context;
  security_context_t context_string;
  int errors = 0;
  int status = 0;

  if (change_symlinks)
    status = lgetfilecon(file, &file_context);
  else 
    status = getfilecon(file, &file_context);

  if ((status < 0) && (errno != ENODATA))
    {
      if (force_silent == 0)
	error (0, errno, "%s", file);
      return 1;
    }

  /* If the file doesn't have a context, and we're not setting all of
     the context components, there isn't really an obvious default.
     Thus, we just give up. */
  if (file_context == NULL && specified_context == NULL)
    {
      error (0, 0, _("can't apply partial context to unlabeled file %s"), file);
      return 1;
    }
      
  if (specified_context == NULL)
    {
      if (compute_context_from_mask (file_context, &context))
	{
	  error (0, 0, _("couldn't compute security context from %s"), file_context);
	      return 1;
	}
    }
  else
    {
      context = context_new (specified_context);
      if (!context)
	error (1, 0,_("invalid context: %s"),specified_context);	    
    }

  context_string = context_str (context);
      
  if (file_context == NULL || strcmp(context_string,file_context)!=0)
    {
      int fail;

      if (change_symlinks)
      	fail = lsetfilecon (file, context_string);
      else
      	fail = setfilecon (file, context_string);

      if (verbosity == V_high || (verbosity == V_changes_only && !fail))
	describe_change (file, context_string, (fail ? CH_FAILED : CH_SUCCEEDED));

      if (fail)
	{
	  errors = 1;
	  if (force_silent == 0)
	    {
	      error (0, errno, _("failed to change context of %s to %s"), file, context_string);
	    }
	}
    }
  else if (verbosity == V_high)
    {
      describe_change (file, context_string, CH_NO_CHANGE_REQUESTED);
    }

  context_free(context);
  freecon(file_context);

  if (recurse) {
    if (lstat(file, &file_stats)==0)
      if (S_ISDIR (file_stats.st_mode))
	errors |= change_dir_context (file, &file_stats);
  }
  return errors;
}

/* Recursively change context of the files in directory DIR
   using specified context components.
   STATP points to the results of lstat on DIR.
   Return 0 if successful, 1 if errors occurred. */

static int
change_dir_context (const char *dir, const struct stat *statp)
{
  char *name_space, *namep;
  char *path;			/* Full path of each entry to process. */
  unsigned dirlength;		/* Length of `dir' and '\0'. */
  unsigned filelength;		/* Length of each pathname to process. */
  unsigned pathlength;		/* Bytes allocated for `path'. */
  int errors = 0;

  errno = 0;
  name_space = savedir (dir);
  if (name_space == NULL)
    {
      if (errno)
        {
	  if (force_silent == 0)
	    error (0, errno, "%s", dir);
	  return 1;
	}
      else
	error (1, 0, _("virtual memory exhausted"));
    }

  dirlength = strlen (dir) + 1;	/* + 1 is for the trailing '/'. */
  pathlength = dirlength + 1;
  /* Give `path' a dummy value; it will be reallocated before first use. */
  path = xmalloc (pathlength);
  strcpy (path, dir);
  path[dirlength - 1] = '/';

  for (namep = name_space; *namep; namep += filelength - dirlength)
    {
      filelength = dirlength + strlen (namep) + 1;
      if (filelength > pathlength)
	{
	  pathlength = filelength * 2;
	  path = xrealloc (path, pathlength);
	}
      strcpy (path + dirlength, namep);
      errors |= change_file_context (path);
    }
  free (path);
  free (name_space);
  return errors;
}

static void
usage (int status)
{
  if (status != 0)
    fprintf (stderr, _("Try `%s --help' for more information.\n"),
	     program_name);
  else
    {
      printf (_("\
Usage: %s [OPTION]... CONTEXT FILE...\n\
  or:  %s [OPTION]... [-u USER] [-r ROLE] [-l RANGE] [-t TYPE] FILE...\n\
  or:  %s [OPTION]... --reference=RFILE FILE...\n\
"),
	program_name, program_name, program_name);
      printf (_("\
Change the security context of each FILE to CONTEXT.\n\
\n\
  -c, --changes          like verbose but report only when a change is made\n\
  -h, --no-dereference   affect symbolic links instead of any referenced file\n\
                         (available only on systems with lchown system call)\n\
  -f, --silent, --quiet  suppress most error messages\n\
      --reference=RFILE  use RFILE's group instead of using a CONTEXT value\n\
  -u, --user=USER        set user USER in the target security context\n\
  -r, --role=ROLE        set role ROLE in the target security context\n\
  -t, --type=TYPE        set type TYPE in the target security context\n\
  -l, --range=RANGE      set range RANGE in the target security context\n\
  -R, --recursive        change files and directories recursively\n\
  -v, --verbose          output a diagnostic for every file processed\n\
      --help             display this help and exit\n\
      --version          output version information and exit\n\
"));
      close_stdout ();
    }
  exit (status);
}

int
main (int argc, char **argv)
{
  security_context_t ref_context = NULL;
  int errors = 0;
  int optc;
  int component_specified = 0;
  
  program_name = argv[0];
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  recurse = force_silent = 0;
  
  while ((optc = getopt_long (argc, argv, "Rcfhvu:r:t:l:", long_options, NULL)) != -1)
  {
	  switch (optc)
	  {
	  case 0:
	    break;
	  case 'u':
	    specified_user = optarg;
	    component_specified = 1;
	    break;
	  case 'r':
	    specified_role = optarg;
	    component_specified = 1;
	    break;
	  case 't':
	    specified_type = optarg;
	    component_specified = 1;
	    break;
	  case 'l':
	    specified_range = optarg;
	    component_specified = 1;
	    break;
	  case CHAR_MAX + 1:
		  reference_file = optarg;
		  break;
	  case 'R':
		  recurse = 1;
		  break;
 	  case 'c':
		  verbosity = V_changes_only;
		  break;
	  case 'f':
		  force_silent = 1;
		  break;
	  case 'h':
		  change_symlinks = 1;
		  break;
	  case 'v':
		  verbosity = V_high;
		  break;
	  default:
		  usage (1);
	  }
  }

  if (show_version)
  {
     printf ("chcon (%s) %s\n", GNU_PACKAGE, VERSION);
     close_stdout ();
     exit (0);
  }

  if (show_help)
    usage (0);

  
  if (reference_file && component_specified)
    {
      error (0, 0, _("conflicting security context specifiers given"));
      usage (1);
    }

  if (!(((reference_file || component_specified)
	 && (argc - optind > 0))
	|| (argc - optind > 1)))
    {
      error (0, 0, _("too few arguments"));
      usage (1);
    }
  
  if (reference_file)
    {
      if (getfilecon (reference_file, &ref_context)<0)
	error (1, errno, "%s", reference_file);
      
      specified_context = ref_context;
    }
  else if (!component_specified) {
    specified_context = argv[optind++];
  }
  for (; optind < argc; ++optind)
     errors |= change_file_context (argv[optind]);

  if (verbosity != V_off)
    close_stdout ();
  if (ref_context != NULL)
    freecon(ref_context);
  exit (errors);
}
