/*
 * runcon [ context |
 *         ( [ -c ] [ -r role ] [-t type] [ -u user ] [ -l levelrange ] )
 *         command [arg1 [arg2 ...] ]
 *
 * attempt to run the specified command with the specified context.
 * 
 * -r role  : use the current context with the specified role
 * -t type  : use the current context with the specified type
 * -u user  : use the current context with the specified user
 * -l level : use the current context with the specified level range
 * -c       : compute process transition context before modifying
 *
 * Contexts are interpreted as follows:
 *
 * Number of       MLS
 * components    system?
 *
 *     1            -         type
 *     2            -         role:type
 *     3            Y         role:type:range
 *     3            N         user:role:type
 *     4            Y         user:role:type:range
 *     4            N         error
 */

#include <config.h>
#include <unistd.h>
#include <stdio.h>
#include <getopt.h>
#include <selinux/context.h>
#include <selinux/selinux.h>
#include <selinux/flask.h>
#include <errno.h>
#include "system.h"
extern int errno;

/* The name the program was run with. */
char *program_name;

/* If nonzero, display usage information and exit.  */
static int show_help;

/* If nonzero, print the version on standard output and exit.  */
static int show_version;

void
usage(int status)
{
  printf(_("Usage: %s CONTEXT COMMAND [args]\n"
         "  or:  %s [ -c ] [-u USER] [-r ROLE] [-t TYPE] [-l RANGE] COMMAND [args]\n"
         "Run a program in a different security context.\n\n"
         "  CONTEXT            Complete security context\n"
         "  -c, --compute      compute process transition context before modifying\n"
         "  -t, --type=TYPE    type (for same role as parent)\n"
         "  -u, --user=USER    user identity\n"
         "  -r, --role=ROLE    role\n"
         "  -l, --range=RANGE  levelrange\n"
         "      --help         display this help and exit\n"
         "      --version      output version information and exit\n"),
         program_name, program_name);
  exit(status);
}

int 
main(int argc,char **argv,char **envp )
{
  char *role    = 0;
  char *range   = 0;
  char *user    = 0;
  char *type    = 0;
  char *context = NULL;
  security_context_t cur_context = NULL;
  security_context_t file_context = NULL;
  security_context_t new_context = NULL;
  int compute_trans = 0;

  context_t      con;

  program_name = argv[0];
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  
  while (1) {
    int c;
    int this_option_optind = optind ? optind : 1;
    int option_index = 0;
    static struct option long_options[] = {
      { "role", 1, 0, 'r' },
      { "type", 1, 0, 't' },
      { "user", 1, 0, 'u' },
      { "range", 1, 0, 'l' },
      { "compute", 0, 0, 'c' },
      { "help", 0, &show_help, 1 },
      { "version", 0, &show_version, 1 },
      { 0, 0, 0, 0 }
    };
    c = getopt_long(argc, argv, "r:t:u:l:c", long_options, &option_index);
    if ( c == -1 ) {
      break;
    }
    switch ( c ) {
    case 0:
      break;
    case 'r':
      if ( role ) {
	fprintf(stderr,_("multiple roles\n"));
	exit(1);
      }
      role = optarg;
      break;
    case 't':
      if ( type ) {
	fprintf(stderr,_("multiple types\n"));
	exit(1);
      }
      type = optarg;
      break;
    case 'u':
      if ( user ) {
	fprintf(stderr,_("multiple users\n"));
	exit(1);
      }
      user = optarg;
      break;
    case 'l':
      if ( range ) {
	fprintf(stderr,_("multiple levelranges\n"));
	exit(1);
      }
      range = optarg;
      break;
    case 'c':
      compute_trans = 1;
      break;
    default:
      usage(1);
      break;
    }
  }

  if (show_version) {
    printf("runcon (%s) %s\n", GNU_PACKAGE, VERSION);
    exit(0);
  }

  if (show_help)
    usage(0);

  if ( !(user || role || type || range || compute_trans)) {
    if ( optind >= argc ) {
      fprintf(stderr,_("must specify -c, -t, -u, -l, -r, or context\n"));
      usage(1);
    }
    context = argv[optind++];
  }
  
  if ( optind >= argc ) {
    fprintf(stderr,_("no command found\n"));
    usage(1);
  }

 if( is_selinux_enabled() != 1 ) {
    fprintf( stderr,
             _("runcon may be used only on a SELinux kernel.\n") );
    exit(-1);
  }
  
  if ( context ) {
    con = context_new(context);
    if (!con) {
      fprintf(stderr,_("%s is not a valid context\n"), context);
      exit(1);
    }
  }
  else {
    if (getcon(&cur_context) < 0) {
      fprintf(stderr,_("Couldn't get current context.\n"));
      exit(1);
    }

    /* We will generate context based on process transition */
    if ( compute_trans ) {
      /* Get context of file to be executed */
      if (getfilecon(argv[optind], &file_context) == -1) {
        fprintf(stderr,_("unable to retrieve attributes of %s\n"),
                argv[optind]);
        exit(1);
      }
      /* compute result of process transition */
      if (security_compute_create(cur_context, file_context,
                                  SECCLASS_PROCESS, &new_context) != 0) {
        fprintf(stderr,_("unable to compute a new context\n"));
        exit(1);
      }
      /* free contexts */
      freecon(file_context);
      freecon(cur_context);

      /* set cur_context equal to new_context */
      cur_context = new_context;
    }

    con = context_new(cur_context);
    if (!con) {
      fprintf(stderr,_("%s is not a valid context\n"), cur_context);
      exit(1);
    }
    if ( user ) {
      if ( context_user_set(con,user)) {
        fprintf(stderr,_("failed to set new user %s\n"),user);
        exit(1);
      }
    }
    if ( type ) {
      if ( context_type_set(con,type)) {
        fprintf(stderr,_("failed to set new type %s\n"),type);
        exit(1);
      }
    }
    if ( range ) {
      if ( context_range_set(con,range)) {
        fprintf(stderr,_("failed to set new range %s\n"),range);
        exit(1);
      }
    }
    if ( role ) {
      if (context_role_set(con,role)) {
        fprintf(stderr,_("failed to set new role %s\n"),role);
        exit(1);
      }
    }
  }

  if (security_check_context(context_str(con)) < 0) {
    fprintf(stderr, _("%s is not a valid context\n"), context_str(con));
    exit(1);
  }
  
  if (setexeccon(context_str(con))!=0) {
    fprintf(stderr,_("unable to setup security context %s\n"), context_str(con));
    exit(1);
  }
  if (cur_context!=NULL) 
    freecon(cur_context);

  if ( execvp(argv[optind],argv+optind) ) {
    perror("execvp");
    exit(1);
  }
  return 1; /* can't reach this statement.... */
}
